#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "read_input.h"
#include "structs.h"
#include "binencode.h"
#include "invdist.h"
#include "condense.h"
#ifdef MPBPA
#include "mpi.h"
#endif

void
check_genes (int *genes, int num_genes, int at_genome)
{
  int i, g0, g1;

  int *checkArr;

  checkArr = (int *) malloc ((num_genes + 1) * sizeof (int));
  if (checkArr == (int *) NULL)
    fprintf (stderr, "ERROR: checkArr NULL\n");

  for (i = 1; i <= num_genes; i++)
    checkArr[i] = 0;

  for (i = 0; i < num_genes; i++)
    {
      g0 = genes[i];
      g1 = (g0 > 0 ? g0 : -g0);
      if (g1 == 0)
	{
	  fflush (outfile);
	  fclose (outfile);
	  fprintf (stderr, "ERROR: Reading Input Genome %d\n", at_genome);
	  fprintf (stderr, "ERROR: Input gene == 0\n");
	  exit (-1);
	}
      if (g1 > num_genes)
	{
	  fflush (outfile);
	  fclose (outfile);
	  fprintf (stderr, "ERROR: Reading Input Genome %d\n", at_genome);
	  fprintf (stderr,
		   "ERROR: Input gene (%d) larger than number of genes %d\n",
		   g0, num_genes);
	  exit (-1);
	}
      if (checkArr[g1] != 0)
	{
	  fflush (outfile);
	  fclose (outfile);
	  fprintf (stderr, "ERROR: Reading Input Genome %d\n", at_genome);
	  fprintf (stderr, "ERROR: Duplicate gene (%d)\n", g1);
	  exit (-1);
	}

      checkArr[g1] = 1;
    }

  for (i = 1; i <= num_genes; i++)
    {
      if (checkArr[i] != 1)
	{
	  fflush (outfile);
	  fclose (outfile);
	  fprintf (stderr, "ERROR: Reading Input Genome %d\n", at_genome);
	  fprintf (stderr, "ERROR: Missing gene (%d)\n", i);
	  exit (-1);
	}
    }

  free (checkArr);
  return;
}

/* Read data file of gene orders in format: 
   >genome name 1 2 4 3 5 -6 -10 -9 -8 7 
   genes 1..n.  */
void
read_data (FILE * input, struct genome_struct **genome_list,
	   int *num_genes, int *num_genomes, int INVDIST, int CIRCULAR,
	   int *condense_succ, int *condense_decode, int *orig_num_genes)
{
  int genome_num;
  int i, j, score;
  int a;
  int g;
  char buf[MAX_STR_LEN];
  char genebuf[MAX_STR_LEN];
  int readingName, continues;
  int gnum;
  int NumGenomes, NumGenes;

  NumGenomes = 0;
  NumGenes = -1;
  /************/
  readingName = 1;
  j = 0;
  a = fgetc (input);
  continues = 1;
  while (continues == 1)
    {

      if (j == MAX_STR_LEN)
	{
	  fflush (outfile);
	  fclose (outfile);
	  fprintf (stderr, "ERROR: Buffer overflow reading input\n");
	  exit (-1);
	}
      switch (a)
	{
	case EOF:
	  continues = 0;
	case '>':
	  if (j > 1)
	    {
	      buf[j] = '\0';
	      j = 0;
	      gnum = strlen (buf);
	      if (gnum <= 1)
		{
		  printf ("Sequence too short\n");
		  exit (-1);
		}
	      while (gnum >= 0 && buf[gnum - 1] == ' ')
		gnum--;
	      if (gnum == 0)
		{
		  printf ("Line is empty\n");
		  exit (-1);
		}
	      if (NumGenes < gnum)
		NumGenes = gnum;
	      readingName = 1 - readingName;
	    }
	  if (a == EOF)
	    {
	      fprintf (outfile, "EOF\n");
	      break;
	    }
	  j = 0;
	  buf[j] = a;
	  j++;
	  break;
	case '\n':
	  if (j > 1)
	    {
	      if (readingName)
		{
		  buf[j] = '\0';
		  j = 0;
		  if (buf[0] != '>')
		    {
		      fflush (outfile);
		      fclose (outfile);
		      fprintf (stderr,
			       "\nERROR: reading genome name from input. ");
		      fprintf (stderr, "Expecting format1: \">Name\" \n");
		      exit (-1);
		    }
		  NumGenomes++;
		  readingName = 1 - readingName;
		}
	    }
	  break;
	default:
	  buf[j] = a;
	  j++;
	}
      a = fgetc (input);
    }

  if (NumGenomes > MAX_GENOMES)
    {
      fprintf (stderr,
	       "ERROR: \tNumber of genomes in the input is %d, but\n",
	       NumGenomes);
      fprintf (stderr,
	       "\tGRAPPA is compiled with a maximum of %d genomes.\n",
	       MAX_GENOMES);
      fprintf (stderr, "\tPlease edit the header file \"structs.h\", ");
      fprintf (stderr, "increase the value\n");
      fprintf (stderr, "\tof MAX_GENOMES (currently set to %d), ",
	       MAX_GENOMES);
      fprintf (stderr, "and recompile.\n");
      exit (-1);
    }

  if (NumGenes > MAX_NUM_GENES)
    {
      fprintf (stderr,
	       "ERROR: \tNumber of genes in the input is %d, but\n",
	       NumGenes);
      fprintf (stderr,
	       "\tGRAPPA is compiled with a maximum of %d genes.\n",
	       MAX_NUM_GENES);
      fprintf (stderr, "\tPlease edit the header file \"structs.h\", ");
      fprintf (stderr, "increase the value\n");
      fprintf (stderr, "\tof MAX_NUM_GENES (currently set to %d), ",
	       MAX_NUM_GENES);
      fprintf (stderr, "and recompile.\n");
      exit (-1);
    }

  rewind (input);

#ifdef MPBPA
  if (MYPROC == 0)
    {
#endif
      fprintf (outfile, "Number of Genomes:\t%d\n", NumGenomes);
      fprintf (outfile, "Number of Genes:  \t%d\n", NumGenes);
      fflush (outfile);
#ifdef MPBPA
    }
#endif

  *genome_list =
    (struct genome_struct *) malloc (NumGenomes *
				     sizeof (struct genome_struct));
  if (*genome_list == (struct genome_struct *) NULL)
    fprintf (stderr, "ERROR: genome_list NULL\n");

  for (i = 0; i < NumGenomes; i++)
    {
      (*genome_list)[i].gnamePtr = (char *) malloc (MAX_NAME * sizeof (char));
      if ((*genome_list)[i].gnamePtr == (char *) NULL)
	{
	  fprintf (stderr, "ERROR: gname NULL\n");
	}
    }

  /************/
  genome_num = 0;
  readingName = 1;
  j = 0;
  a = fgetc (input);
  continues = 1;
  while (continues == 1)
    {

      if (j == MAX_STR_LEN)
	{
	  fprintf (stderr, "ERROR: Buffer overflow reading input\n");
	  exit (-1);
	}
      switch (a)
	{
	case EOF:
	  continues = 0;
	case '>':
	  if (j > 1)
	    {
	      buf[j] = '\0';
	      j = 0;
	      gnum = strlen (buf);
	      while (gnum >= 0 &&
		     (buf[gnum - 1] == ' ' || buf[gnum - 1] == '\n'))
		gnum--;
	      /* get rid of the ending white space */

	      for (i = 0; i < gnum; i++)
		(*genome_list)[genome_num].genes[i] = buf[i];

	      (*genome_list)[genome_num].num_chars = gnum;
	      /*check_genes ( ( *genome_list )[genome_num].genes,
	         NumGenes, genome_num + 1 ); */
	      genome_num++;
	      readingName = 1 - readingName;
	    }
	  if (a == EOF)
	    {
	      fprintf (outfile, "EOF\n");
	      break;
	    }
	  j = 0;
	  buf[j] = a;
	  j++;
	  break;
	case '\n':
	  if (j > 1)
	    {
	      if (readingName)
		{
		  buf[j] = '\0';
		  j = 0;
		  if (buf[0] != '>')
		    {
		      fprintf (stderr,
			       "\nERROR: reading genome name from input. ");
		      fprintf (stderr, "Expecting format2: \">Name\" \n");
		      exit (-1);
		    }
		  if (strlen (buf + 1) > MAX_STR_LEN)
		    fprintf (stderr,
			     "ERROR: Buffer overflow in genome name\n");
		  strcpy ((*genome_list)[genome_num].gnamePtr, buf + 1);
		  (*genome_list)[genome_num].genome_num = genome_num + 1;
		  (*genome_list)[genome_num].genes =
		    (char *) malloc ((2 * NumGenes + 2) * sizeof (int));
		  if ((*genome_list)[genome_num].genes == NULL)
		    fprintf (stderr, "ERROR: cannot allocate genes\n");
		  readingName = 1 - readingName;
		}
	    }
	  break;
	default:
	  buf[j] = a;
	  j++;
	}
      a = fgetc (input);
    }

  fclose (input);

#ifdef MPBPA
  if (MYPROC == 0)
    {
#endif
/* STACIA REDUCING OUTPUT 
    printGenomes(*genome_list, NumGenomes, NumGenes);
    fprintf(outfile,"\n");
*/
#ifdef MPBPA
    }
#endif

  *orig_num_genes = NumGenes;

  *num_genes = NumGenes;
  *num_genomes = NumGenomes;

  printf ("%d\n", *num_genomes);
  for (i = 0; i < *num_genomes; i++)
    {
      for (j = 0; j < (*genome_list)[i].num_chars; j++)
	{
	  printf ("%c", (char) (*genome_list)[i].genes[j]);
	}
      printf ("\n");
    }
  return;
}
